from aiotfm.errors import AiotfmException


class Room:
	"""Represents the room that the bot currently is in.

	Attributes
	----------
	name: `str`
		The room's name. (i.e: en-1, *bad girls and so on)
	private: `bool`
		Whether the room is public or private.
	players: `list[aiotfm.player.Player]`
		The list containing all the players of the room.
	"""
	def __init__(self, name, private=True):
		self.name = name
		self.private = private
		self.players = {}

	def __repr__(self):
		return "<Room name={} private={}>".format(self.name, self.private)

	@property
	def community(self):
		"""Returns the room's community."""
		if self.name.startswith('*'):
			return 'xx'
		return self.name.split('-', 1)[0]

	@property
	def is_tribe(self):
		"""Returns true if it's a tribe house."""
		return self.name.startswith('*\x03')

	@property
	def display_name(self):
		"""Return the display name of the room.
		It removes the \x03 char from the tribe house and the community from the public rooms."""
		if self.is_tribe:
			return self.name.replace('\x03', '')
		if self.name.startswith('*'):
			return self.name
		return self.name.split('-', 1)[1]

	def get_players(self, predicate, max_=None):
		"""Filters players from the room.

		:param predicate: A function that returns a boolean-like result to filter through
			the players.
		:param max_: Optional[:class:`int`] The maximum amount of players to return.
		:return: `Iterable` The filtered players."""
		return [p for p in self.players.values() if predicate(p)][:max_]

	def get_player(self, default=None, **kwargs):
		"""Gets one player in the room with an identifier.

		:param kwargs: Which identifier to use. Can be either name, username, id or pid.
		:return: :class:`aiotfm.player.Player` The player or None"""
		length = len(kwargs.keys())

		if length == 0:
			raise AiotfmException('You did not provide any identifier.')
		if length > 1:
			raise AiotfmException('You cannot filter one player with more than one identifier.')

		identifier, value = next(iter(kwargs.items()))

		if identifier in ('name', 'username'):
			def filter_(p):
				return p == value
		elif identifier == 'id':
			def filter_(p):
				return p.id == int(value)
		elif identifier == 'pid':
			return self.players.get(int(value), default)
		else:
			raise AiotfmException('Invalid filter.')

		for player in self.players.values():
			if filter_(player):
				return player
		return default


from aiotfm.packet import Packet


class Shop:
	"""Represents the shop in game.

	Parameters
	----------
	packet: :class:`aiotfm.Packet`
		The packet where the shop content cill be read.

	Attributes
	----------
	cheese: :class:`int`
		The number of cheese the client has.
	fraise: :class:`int`
		The numberof fraise (strawberries) the client has.
	look: :class:`str`
		The client's look.
	owned_items: :class:`set`[:class:`aiotfm.shop.Item`]
		All items the client own.
	items: :class:`set`[:class:`aiotfm.shop.ShopItem`]
		All items present in the shop.
	full_outfits: :class:`set`[:class:`aiotfm.shop.Outfit`]
		Available fashion outfits you can buy.
	outfits: :class:`set`[:class:`aiotfm.shop.Outfit`]
		The client own outfits.
	owned_shaman_objects: :class:`set`[:class:`aiotfm.shop.OwnedShamanObject`]
		All shaman object the client own.
	shaman_objects: :class:`set`[:class:`aiotfm.shop.ShamanObject`]
		All shaman object available in the shop.
	"""
	def __init__(self, packet: Packet):
		self.cheese = packet.read32()
		self.fraise = packet.read32()
		self.look = packet.readUTF()

		self.owned_items = set(Item.from_packet(packet) for _ in range(packet.read32()))
		self.items = set(ShopItem.from_packet(packet) for _ in range(packet.read32()))

		self.full_outfits = set(Outfit.from_fashion(packet) for _ in range(packet.read8()))
		self.outfits = set(Outfit.from_packet(packet, i) for i in range(packet.read16()))

		self.owned_shaman_objects = set(
			OwnedShamanObject.from_packet(packet) for _ in range(packet.read16())
		)
		self.shaman_objects = set(ShamanObject.from_packet(packet) for _ in range(packet.read16()))

	def to_dict(self):
		"""Export the shop into a serializable dict.
		:return: :class:`dict`
		"""
		items = [item.to_dict() for item in sorted(self.items, key=lambda i: i.uid)]

		return {
			"cheese": self.cheese,
			"fraise": self.fraise,
			"look": self.look,
			"items": items
		}

	def cost(self, outfit):
		"""Compute and return the total price of an outfit.
		:param outfit: :class:`aiotfm.shop.Outfit`
		:return: :class:`tuple`[:class:`int`]
			(Total in cheese, total in fraise, cheese supplement of items you can't buy with fraise)
		"""
		cheese = 0
		fraise = 0
		fraise_sup = 0
		for item in outfit.items:
			if item.id == 0:
				continue
			item = self.getItem(item)
			cheese += item.cheese if item.cheese < 1000001 else 0
			if item.fraise == 0 and item.cheese < 1000001:
				fraise_sup += item.cheese
			else:
				fraise += item.fraise

		if outfit.fur != 1:
			fur = [i for i in self.items if i.id == outfit.fur and (21 <= i.category <= 22)][0]
			cheese += fur.cheese if fur.cheese < 1000001 else 0
			fraise += fur.fraise
			if fur.fraise == 0 and fur.cheese < 1000001:
				fraise_sup += fur.cheese

		return cheese, fraise, fraise_sup

	def getItem(self, item):
		"""Return the shop item with the same id.
		:param item: :class:`aiotfm.shop.Item` the item you want to price of.
		:return: :class:`aiotfm.shop.ShopItem` the item with the prices.
		"""
		for i in self.items:
			if i == item:
				return i
		return None

	def category(self, id_):
		"""Return the items from a category.
		:param id_: :class:`int` the category's id?
		:return: :class:`set`[:class:`aiotfm.shop.Item`] the items.
		"""
		return set(i for i in self.items if i.category == id_)


class Item:
	"""Represents an item from the shop.

	Parameters
	----------
	category: :class:`int`
		The item's category.
	id_: :class:`int`
		The item's id.
	colors: Optional[:class:`int`]
		The item's colors.

	Attributes
	----------
	category: :class:`int`
		The item's category.
	id: :class:`int`
		The item's id.
	uid: :class:`int`
		The item's unique id.
	colors: :class:`int`
		The item's colors.
	"""
	def __init__(self, category, id_, colors=None):
		self.category, self.id = int(category), int(id_)
		self.uid = self.category * 10000 + self.id

		self.colors = colors or []

	def __eq__(self, other):
		return self.category == other.category and self.id == other.id

	def __hash__(self):
		return self.uid

	@classmethod
	def from_packet(cls, packet: Packet):
		"""Reads an Item from a packet.
		:param packet: :class:`aiotfm.Packet`
		:return: :class:`aiotfm.shop.Item`
		"""
		nbr_colors = packet.read8()
		uid = packet.read32()
		cat = (uid - 10000) // 10000 if uid > 9999 else uid // 100

		if uid < 99:
			id_ = uid
		elif uid < 999:
			id_ = uid % (100 * cat)
		elif uid < 9999:
			id_ = uid % 100
		else:
			id_ = uid % 1000

		colors = []
		if nbr_colors > 0:
			colors = [packet.read32() for i in range(nbr_colors - 1)]

		return cls(cat, id_, colors)

	@classmethod
	def parse(cls, cat, string):
		"""Parse an Item from a string.
		:param cat: :class:`int` the item's category.
		:param string: :class:`str` the item.
		:return: :class:`aiotfm.shop.Item`
		"""
		string = string.split('_')
		id_ = int(string[0])
		if len(string) == 1:
			return cls(cat, id_)
		return cls(cat, id_, [int(color, 16) for color in string[1].split('+')])


class ShopItem(Item):
	"""Represents an item from the shop with its specifications.

	Parameters
	----------
	category: :class:`int`
		The item's category.
	id_: :class:`int`
		The item's id.
	colors: :class:`int`
		The item's colors.
	is_new: :class:`bool`
		True if it's a new item.
	flag: :class:`int`
		Contains the item's metadata.
	cheese: :class:`int`
		The item's price in cheese.
	fraise: :class:`int`
		The item's price in fraise.
	special: :class:`int`
		The item's special data.

	Attributes
	----------
	category: :class:`int`
		The item's category.
	id: :class:`int`
		The item's id.
	uid: :class:`int`
		The item's unique id.
	colors: :class:`int`
		The item's colors.
	nbr_colors: :class:`int`
		The number of customizable colors the item has.
	is_new: :class:`bool`
		True if it's a new item.
	flag: :class:`int`
		Contains the item's metadata.
	cheese: :class:`int`
		The item's price in cheese.
	fraise: :class:`int`
		The item's price in fraise.
	special: :class:`int`
		The item's special data.
	"""
	def __init__(self, category, id_, colors, is_new, flag, cheese, fraise, special):
		super().__init__(category, id_)

		self.nbr_colors = colors
		self.is_new = is_new
		self.flag = flag
		self.cheese = cheese
		self.fraise = fraise
		self.special = special
		if self.is_new:
			print(category, id_, colors, is_new, flag, cheese, fraise, special)

	def to_dict(self):
		"""Export the item into a serializable dict.
		:return: :class:`dict`
		"""
		data = {
			'category': self.category,
			'cheese': self.cheese,
			'colors': self.nbr_colors,
			'fraise': self.fraise,
			'id': self.id
		}
		if self.cheese == 1000001:
			data['purchasable'] = False
			data['cheese'] = 0
		if self.flag == 13:
			data['collector'] = self.flag == 13
		if self.is_new:
			data['new'] = self.is_new
		return data

	@classmethod
	def from_packet(cls, packet: Packet):
		"""Reads a ShopItem from a packet.
		:param packet: :class:`aiotfm.Packet`
		:return: :class:`aiotfm.shop.ShopItem`
		"""
		return cls(
			packet.read16(), packet.read16(), packet.read8(), packet.readBool(),
			packet.read8(), packet.read32(), packet.read32(), packet.read16()
		)


class Outfit:
	"""Represents an outfit from the shop.

	Parameters
	----------
	look: :class:`str`
		The outfit's look.
	id_: Optional[:class:`int`]
		The outfit's id.
	flag: :class:`int`
		Contains the outfit's metadata.

	Attributes
	----------
	look: :class:`str`
		The outfit's look.
	id: :class:`int`
		The outfit's id.
	flag: :class:`int`
		Contains the outfit's metadata.

	"""
	def __init__(self, look, id_=-1, flag=-1):
		self.look = look
		self.id = id_
		self.flag = flag

	def __eq__(self, other):
		if isinstance(other, str):
			return self.look == other
		return self.look == other.look

	def __hash__(self):
		return hash(hash(self.look) + hash(self.id))

	@classmethod
	def from_fashion(cls, packet: Packet):
		"""Reads a fashion Outfit from a packet.
		:param packet: :class:`aiotfm.Packet`
		:return: :class:`aiotfm.shop.Outfit`
		"""
		id_ = packet.read16()
		look = packet.readUTF()
		flag = packet.read8()
		return cls(look, id_, flag)

	@classmethod
	def from_packet(cls, packet: Packet, id_):
		"""Reads an Outfit from a packet.
		:param packet: :class:`aiotfm.Packet`
		:return: :class:`aiotfm.shop.Outfit`
		"""
		return cls(packet.readUTF(), id_)

	@property
	def fur(self):
		"""The fur's id of the outfit."""
		return int(self.look.split(';')[0])

	@property
	def items(self):
		"""The outfit's items."""
		items = []
		for i, item in enumerate(self.look.split(';')[1].split(',')):
			items.append(Item.parse(i, item))
		return items

	@property
	def head(self):
		"""The outfit's head item."""
		return self.items[0]

	@property
	def eyes(self):
		"""The outfit's eyes item."""
		return self.items[1]

	@property
	def ears(self):
		"""The outfit's ears item."""
		return self.items[2]

	@property
	def mouth(self):
		"""The outfit's mouth item."""
		return self.items[3]

	@property
	def neck(self):
		"""The outfit's neck item."""
		return self.items[4]

	@property
	def hair(self):
		"""The outfit's hair item."""
		return self.items[5]

	@property
	def tail(self):
		"""The outfit's tail item."""
		return self.items[6]

	@property
	def lenses(self):
		"""The outfit's lenses item."""
		return self.items[7]

	@property
	def hands(self):
		"""The outfit's hands item."""
		return self.items[8]


class ShamanObject:
	"""Represents shaman object from the shop.

	Parameters
	----------
	id_: :class:`int`
		The object's id.
	colors: :class:`int`
		The number of customizable colors the object has.
	is_new: :class:`bool`
		The object's metadata.
	flag: :class:`int`
		Contains the object's metadata.
	cheese: :class:`int`
		The obect's pricein cheese.
	fraise: :class:`int`
		The obect's pricein fraise.

	Attributes
	----------
	id: :class:`int`
		The object's id.
	colors: :class:`int`
		The number of customizable colors the object has.
	is_new: :class:`bool`
		The object's metadata.
	flag: :class:`int`
		Contains the object's metadata.
	cheese: :class:`int`
		The obect's pricein cheese.
	fraise: :class:`int`
		The obect's pricein fraise.

	"""
	def __init__(self, id_, colors, is_new, flag, cheese, fraise):
		self.id = id_
		self.colors = colors
		self.is_new = is_new
		self.flag = flag
		self.cheese = cheese
		self.fraise = fraise

	def __eq__(self, other):
		return self.id == other.id

	def __hash__(self):
		return self.id

	@classmethod
	def from_packet(cls, packet: Packet):
		"""Reads a ShamanObject from a packet.
		:param packet: :class:`aiotfm.Packet`
		:return: :class:`aiotfm.shop.ShamanObject`
		"""
		return cls(
			packet.read32(), packet.read8(), packet.readBool(), packet.read8(),
			packet.read32(), packet.read16()
		)


class OwnedShamanObject:
	"""Represents shaman object that the client own.

	Parameters
	----------
	id_: :class:`int`
		The object's id.
	equiped: :class:`bool`
		True if the client has the object equiped.
	colors: :class:`list`[:class:`int`]
		The custom colors the object has.

	Attributes
	----------
	id: :class:`int`
		The object's id.
	equiped: :class:`bool`
		True if the client has the object equiped.
	colors: :class:`list`[:class:`int`]
		The custom colors the object has.

	"""
	def __init__(self, id_, equiped, colors):
		self.id = id_
		self.equiped = equiped
		self.colors = colors

	def __eq__(self, other):
		return self.id == other.id

	def __hash__(self):
		return self.id

	@classmethod
	def from_packet(cls, packet: Packet):
		"""Reads a OwnedShamanObject from a packet.
		:param packet: :class:`aiotfm.Packet`
		:return: :class:`aiotfm.shop.OwnedShamanObject`
		"""
		id_ = packet.read16()
		equiped = packet.readBool()
		colors = [packet.read32() for i in range(max(packet.read8() - 1, 0))]

		return cls(id_, equiped, colors)


from enum import IntEnum


class ChatCommunity(IntEnum):
	"""Enumerate the different chat's communities."""
	int = -1
	en = 1
	fr = 2
	ru = 3
	br = 4
	es = 5
	cn = 6
	tr = 7
	vk = 8
	pl = 9
	hu = 10
	nl = 11
	ro = 12
	id = 13
	de = 14
	e2 = 15
	ar = 16
	ph = 17
	lt = 18
	jp = 19
	fi = 21
	cz = 22
	hr = 23
	bg = 25
	lv = 26
	he = 27
	it = 28
	pt = 31

	@classmethod
	def _missing_(cls, value):
		return cls.int


class Community(IntEnum):
	"""Enumerate the different game's communities."""
	en = 0
	int = 0
	fr = 1
	ru = 2
	br = 3
	es = 4
	cn = 5
	tr = 6
	vk = 7
	pl = 8
	hu = 9
	nl = 10
	ro = 11
	id = 12
	de = 13
	e2 = 14
	ar = 15
	ph = 16
	lt = 17
	jp = 18
	fi = 20
	cz = 21
	hr = 23
	bg = 24
	lv = 25
	he = 26
	it = 27
	et = 29
	pt = 31

	@classmethod
	def _missing_(cls, value):
		return cls.int


class TradeState(IntEnum):
	"""Enumerate the different states a trade can have."""
	ON_INVITE = 0
	ACCEPTING = 1
	TRADING = 2
	CANCELLED = 3
	SUCCESS = 4


class TradeError(IntEnum):
	"""Enumerate the different error codes the server could send."""
	ALREADY_TRADING = 0
	INVITE_DECLINED = 1
	CANCELLED = 2
	SAME_ROOM = 3
	SUCCEED = 4
	SHAMAN = 5
	NOT_CONNECTED = 6
	INTERNAL = 7


class Permissions(IntEnum):
	"""Enumerate the different tribe's rank's permissions."""
	IS_LEADER = 2
	CAN_CHANGE_GREETING_MESSAGE = 4
	CAN_EDIT_RANKS = 8
	CAN_CHANGE_MEMBERS_RANKS = 16
	CAN_INVITE = 32
	CAN_EXCLUDE = 64
	CAN_PLAY_MUSIC = 128
	CAN_CHANGE_TRIBE_HOUSE_MAP = 256
	CAN_LOAD_MAP = 512
	CAN_LOAD_LUA = 512
	CAN_MANAGE_FORUM = 1024


from aiotfm.utils import Date
from aiotfm.enums import Permissions


class Tribe:
	"""Represents a tribe.

	## Attributes

	id :class:`int` the tribe's id.
	name :class:`str` the tribe's name.
	welcomeMessage :class:`str` the tribe's welcome message.
	mapcode :class:`int` the tribehouse's mapcode.
	members :class:`list` the members' list of the tribe.
	ranks :class:`list` the ranks' list of the tribe.
	"""
	def __init__(self, packet):
		self.id = packet.read32()
		self.name = packet.readUTF()
		self.welcomeMessage = packet.readUTF()
		self.mapcode = packet.read32()
		self.members = []
		self.ranks = []

		for i in range(packet.read16()):
			self.members.append(Member(self, packet))

		for i in range(packet.read16()):
			self.ranks.append(Rank.from_packet(i, packet))


class Member:
	"""Represents a tribe's member.

	## Attributes

	tribe :class:`Tribe` the member's tribe.
	id :class:`int` the player's id of the member.
	name :class:`str` the username of the member.
	gender :class:`int` the member's gender.
	lastConnection :class`Date` the last connection of the member.
	rank_id :class:`int` the rank's id of the member.
	game_id :class:`int` the game's id the player is playing.
	room :class:`str` the room where the player is.
	rank :class:`Rank` the member's rank.
	online :class:`bool` return True if the member is online.
	"""
	def __init__(self, tribe, packet):
		self.tribe = tribe
		self.id = packet.read32()
		self.name = packet.readUTF()
		self.gender = packet.read8()
		packet.read32() # id
		self.lastConnection = Date.fromtimestamp(packet.read32())
		self.rank_id = packet.read8()
		self.game_id = packet.read32()
		self.room = packet.readUTF()

	@property
	def rank(self):
		"""return the :class:`Rank` of the member."""
		return self.tribe.ranks[self.rank_id]

	@property
	def online(self):
		"""return True if the member is online."""
		return self.game_id != 1


class Rank:
	"""Represents a tribe's rank.

	## Attributes

	id :class:`int` the rank's id.
	name :class:`str` the rank's name.
	perm :class:`int` the rank's permissions.

	## Permissions

	isLeader :class:`bool`
	canChangeGreetingMessage :class:`bool`
	canEditRanks :class:`bool`
	canChangeMembersRanks :class:`bool`
	canInvite :class:`bool`
	canExclude :class:`bool`
	canPlayMusic :class:`bool`
	canChangeTribeHouseMap :class:`bool`
	canLoadMap :class:`bool`
	canLoadLua :class:`bool`
	canManageForum :class:`bool`
	"""
	def __init__(self, id_, name, perm):
		self.id = id_
		self.name = name
		self.perm = perm

	@property
	def isLeader(self):
		"""True if it's the tribe's leader's rank."""
		return bool(self.perm & Permissions.IS_LEADER)

	@property
	def canChangeGreetingMessage(self):
		"""True if it has the permission to change the greeting message."""
		return bool(self.perm & Permissions.CAN_CHANGE_GREETING_MESSAGE)

	@property
	def canEditRanks(self):
		"""True if it has the permission to edit ranks."""
		return bool(self.perm & Permissions.CAN_EDIT_RANKS)

	@property
	def canChangeMembersRanks(self):
		"""True if it has the permission to change members' rank."""
		return bool(self.perm & Permissions.CAN_CHANGE_MEMBERS_RANKS)

	@property
	def canInvite(self):
		"""True if it has the permission to invite someone to the tribe."""
		return bool(self.perm & Permissions.CAN_INVITE)

	@property
	def canExclude(self):
		"""True if it has the permission to exclude someone of the tribe."""
		return bool(self.perm & Permissions.CAN_EXCLUDE)

	@property
	def canPlayMusic(self):
		"""True if it has the permission to play music inside the tribe's house."""
		return bool(self.perm & Permissions.CAN_PLAY_MUSIC)

	@property
	def canChangeTribeHouseMap(self):
		"""True if it has the permission to change the tribe's house's map."""
		return bool(self.perm & Permissions.CAN_CHANGE_TRIBE_HOUSE_MAP)

	@property
	def canLoadMap(self):
		"""True if it has the permission to load maps inside the tribe's house."""
		return bool(self.perm & Permissions.CAN_LOAD_MAP)

	@property
	def canLoadLua(self):
		"""True if it has the permission to load Lua inside the tribe's house."""
		return bool(self.perm & Permissions.CAN_LOAD_LUA)

	@property
	def canManageForum(self):
		"""True if it has the permission to mange the tribe's forum."""
		return bool(self.perm & Permissions.CAN_MANAGE_FORUM)

	@classmethod
	def from_packet(cls, id_, packet):
		"""Reads a Tribe from a packet.
		:param id: :class:`int` the tribe's id.
		:param packet: :class:`aiotfm.Packet`"""
		return cls(id_, packet.readUTF(), packet.read32())


from aiotfm import utils, errors, enums
from aiotfm.client import Client
from aiotfm.connection import Connection
from aiotfm.packet import Packet
from aiotfm.player import Player, Profile, Stats
from aiotfm.tribe import Tribe, Member, Rank

from aiotfm.__version__ import __author__, __title__, __description__
from aiotfm.__version__ import __url__, __version__, __credits__, __license__

__all__ = [
	'__author__', '__credits__', '__description__', '__license__', '__title__', '__url__',
	'__version__', 'enums', 'errors', 'utils', 'Client', 'Connection', 'Member', 'Packet',
	'Player', 'Profile', 'Rank', 'Stats', 'Tribe'
]


import asyncio

from functools import cmp_to_key

from aiotfm.packet import Packet
from aiotfm.player import Player
from aiotfm.enums import TradeState
from aiotfm.errors import TradeOnWrongState


class InventoryItem:
	"""Represents an inventory item.

	Attributes
	----------
	id: `int`
		The item id.
	quantity: `int`
		The quantity of the item.
	inventory: Optional[`aiotfm.inventory.Inventory`]
		The inventory class. Might be None.
	can_use: `bool`
		True if you can use this item.
	category: `int`
		Define the category's item. Used by the sorting algorithm.
	img_id: `str`
		Id used to get the item's image.
	is_event: `bool`
		True if it's an item from an event.
	slot: `int`
		Define the equipped slot with this item. If slot is 0 then the item is not equipped.
	"""
	def __init__(self, item_id, **kwargs):
		self.id = item_id
		self.quantity = kwargs.get("quantity", 0)
		self.inventory = kwargs.get("inventory", None)

		self.can_use = kwargs.get("can_use", True)
		self.category = kwargs.get("category", 0)
		self.img_id = kwargs.get("img_id", str(self.id))
		self.is_event = kwargs.get("is_event", False)
		self.slot = kwargs.get("slot", 0)

	def __repr__(self):
		return "<InventoryItem id={} quantity={}>".format(self.id, self.quantity)

	def __eq__(self, other):
		return self.id == other.id

	@property
	def image_url(self):
		"""The image's url of the item."""
		url = 'https://www.transformice.com/images/x_transformice/x_inventaire/{.img_id}.jpg'
		return url.format(self)

	@property
	def is_currency(self):
		"""Return True if the item is a currency."""
		return self.id in (800, 801, 2253, 2254, 2257, 2260, 2261)

	@property
	def is_equipped(self):
		"""Return True if the item is equipped"""
		return self.slot > 0

	@classmethod
	def from_packet(cls, packet):
		"""Read an item from a packet.
		:param packet: :class:`aiotfm.Packet` the packet.
		:return: :class:`aiotfm.inventory.InventoryItem` the item.
		"""
		item_id = packet.read16()
		kwargs = {
			'quantity': packet.read8(),
			'category': packet.read8(),
			'is_event': packet.readBool(),
			'can_use': packet.readBool()
		}
		packet.readBool() # similar to `can_use`
		packet.readBool() # similar to `can_use`
		packet.readBool()
		packet.readBool()
		if packet.readBool():
			kwargs['img_id'] = packet.readUTF()

		# if equipped, this is the slot (1, 2, 3); otherwise this is 0
		kwargs['slot'] = packet.read8()
		return cls(item_id, **kwargs)

	async def use(self):
		"""|coro|
		Uses this item."""
		if self.inventory is None or self.inventory.client is None:
			message = "InventoryItem doesn't have the inventory variable ""or Inventory doesn't \
				have the client variable."
			raise TypeError(message)
		await self.inventory.client.main.send(Packet.new(31, 3).write16(self.id))


class Inventory:
	"""Represents the client's inventory.

	Attributes
	----------
	items: `dict`
		A dict containing all the items. The key is an :class:`int` and the value is
		an :class:`aiotfm.inventory.InventoryItem`.
	client: `aiotfm.client.Client`
		The client that this inventory belongs to.
	"""
	def __init__(self, client=None, items=None):
		self.items = items or {}
		self.client = client

		for item in self:
			item.inventory = self

	def __repr__(self):
		return "<Inventory client={!r}>".format(self.client)

	def __iter__(self):
		return iter(self.items.values())

	def __getitem__(self, index):
		if not isinstance(index, int):
			raise TypeError("Index must be int, not {}".format(type(index)))
		return self.items[index]

	def __setitem__(self, index, value):
		if not isinstance(index, int):
			raise TypeError("Index must be int, not {}".format(type(index)))
		self.items[index] = value

	@classmethod
	def from_packet(cls, packet):
		"""Read the inventory from a packet.
		:param packet: :class:`aiotfm.Packet` the packet.
		:return: :class:`aiotfm.inventory.Inventory` the inventory.
		"""
		items = {}

		for item in range(packet.read16()):
			item = InventoryItem.from_packet(packet)
			items[item.id] = item

		return cls(items=items)

	def get(self, item_id):
		"""Gets an item from this :class:`aiotfm.inventory.Inventory`.
		Shorthand for :class:`aiotfm.inventory.Inventory`.items.get"""
		return self.items.get(item_id, InventoryItem(item_id))

	def sort(self):
		"""Sort the inventory the same way the client does.
		:return: :class:`list`
		"""
		def cmp(a, b):
			if (a.is_currency or b.is_currency) and not (a.is_currency and b.is_currency):
				return -1 if a.is_currency else 1 # Currency are always on the top
			if (a.is_event or b.is_event) and not (a.is_event and b.is_event):
				return -1 if a.is_event else 1 # Event items comes always after the currency
			if a.category != b.category:
				return b.category - a.category # Higher means first
			return a.id - b.id # Lastly the items are sorted by their ids

		return sorted(iter(self), key=cmp_to_key(cmp))


class TradeContainer:
	"""Represents the content of a Trade."""
	def __init__(self, trade):
		self.trade = trade
		self._content = []

	def get(self, item_id, default=0):
		"""Returns the quantity of an item inside the TradeContainer.
		:param item_id: :class:`int` the item's id.
		:param default: Optional[:class:`int`] the default value if the item is not present.
		:return: :class:`int` the quantity of the item.
		"""
		for item in self._content:
			if item.id == item_id:
				return item.quantity
		return default

	def getSlot(self, index):
		"""Returns the item inside a certain slot.
		:param index: :class:`int` the index.
		:return: :class:`aiotfm.inventory.InventoryItem` the item.
		"""
		return self._content[index]

	def add(self, item_id, quantity):
		"""Add a quantity of an item inside the container.
		:param item_id: :class:`int` the item's id.
		:param quantity: :class:`int` the quantity to add. Can be negative.
		"""
		for item in self._content:
			if item.id == item_id:
				item.quantity += quantity
				if item.quantity == 0:
					self._content.remove(item)
				break
		else:
			self._content.append(InventoryItem(item_id, quantity=quantity))


class Trade:
	"""Represents a trade that the client is participating (not started, in progress or ended).

	Attributes
	----------
	client: :class:`aiotfm.Client`
		The reference to the client involved in the trade.
	trader: `str`
		The player the client is trading with.
	locked: List[`bool`]
		A list of two `bool` describing the locked state of each party.
	imports: :class:`aiotfm.inventory.TradeContainer`
		The container of the items you will receive if the trade succeed.
	exports: :class:`aiotfm.inventory.TradeContainer`
		The container of the items you will give if the trade succeed.
	state: :class:`aiotfm.utils.TradeState`
		The current state of the trade.
			ON_INVITE: an invitation has been received from/sent to the other party.
			ACCEPTING: the client accepted and is waiting for the other party to be ready.
			TRADING:   the only state of the trade where you are able to add items.
			CANCELLED: the trade has been cancelled by one of the parties.
			SUCCESS:   the trade finished successfully."""
	def __init__(self, client, trader):
		self.client = client
		self.trader = trader
		self.locked = [False, False] # 0: trader, 1: client

		self.imports = TradeContainer(self)
		self.exports = TradeContainer(self)

		self.state = TradeState.ON_INVITE
		self.pid = -1

		if isinstance(trader, str):
			trader = client.room.get_player(name=trader)
			if trader is None:
				TypeError(f"Can not find the player '{self.trader}' in the room.")

		if isinstance(trader, Player):
			if self.trader.isGuest:
				raise TypeError("You can not trade with a guest.")
			if self.trader == self.client.username:
				raise TypeError("You can not trade with yourself.")
			if self.trader.pid == 0:
				raise TypeError("You can not trade with a player having the same IP.")
			self.trader = self.trader.username
			self.pid = trader.pid
		else:
			raise TypeError(f"Trade expected 'Player' or 'str' type, got '{type(trader)}")

	def __repr__(self):
		return "<Trade state={} locked=[trader:{}, client:{}] trader={} pid={}>".format(
			TradeState[self.state], *self.locked, self.trader, self.pid)

	def __eq__(self, other):
		if other is None:
			return False
		if self.pid == -1 or other.pid == -1:
			return self.trader.lower() == other.trader.lower()
		return self.pid == other.pid

	@property
	def closed(self):
		"""Returns True if the trade is closed."""
		return self.state in (TradeState.SUCCESS, TradeState.CANCELLED)

	def _start(self):
		"""Set the state of the trade as TRADING."""
		self.state = TradeState.TRADING

	def _close(self, succeed=False):
		"""Closes the trade."""
		self.state = TradeState.SUCCESS if succeed else TradeState.CANCELLED
		if self.client.trade == self:
			self.client.trade = None

		self.client.trades.pop(self.pid, None)
		self.client.dispatch('trade_close', self, succeed)

	async def cancel(self):
		"""|coro|
		Cancels the trade."""
		if self.state != TradeState.TRADING:
			raise TradeOnWrongState('cancel', self.state)

		await self.client.main.send(Packet.new(31, 6).writeString(self.trader).write8(2))

	async def accept(self):
		"""|coro|
		Accepts the trade."""
		if self.state != TradeState.ON_INVITE:
			raise TradeOnWrongState('accept', self.state)

		self.state = TradeState.ACCEPTING
		await self.client.main.send(Packet.new(31, 5).writeString(self.trader))

	async def addItem(self, item_id, quantity):
		"""|coro|
		Adds an item to the trade.

		:param item_id: :class:`int` The item id.
		:param quantity: :class:`int` The quantity of item to add."""
		if self.state != TradeState.TRADING:
			raise TradeOnWrongState('addItem', self.state)

		quantity = min(max(quantity, 0), 200)
		packet = Packet.new(31, 8).write16(item_id).writeBool(True).buffer

		ten = packet + b'\x01'
		for i in range(quantity // 10):
			await self.client.main.send(Packet(ten))
			await asyncio.sleep(.05)

		unit = packet + b'\x00'
		for i in range(quantity % 10):
			await self.client.main.send(Packet(unit))
			await asyncio.sleep(.05)

	async def removeItem(self, item_id, quantity):
		"""|coro|
		Removes an item from the trade.

		:param item_id: :class:`int` The item id.
		:param quantity: :class:`int` The quantity of item to remove."""
		if self.state != TradeState.TRADING:
			raise TradeOnWrongState('removeItem', self.state)

		quantity = min(max(quantity, 0), 200)
		packet = Packet.new(31, 8).write16(item_id).writeBool(False).buffer

		ten = packet + b'\x01'
		for i in range(quantity // 10):
			await self.client.main.send(Packet(ten))
			await asyncio.sleep(.05)

		unit = packet + b'\x00'
		for i in range(quantity % 10):
			await self.client.main.send(Packet(unit))
			await asyncio.sleep(.05)

	async def lock(self):
		"""|coro|
		Locks (confirms) the trade."""
		if self.state != TradeState.TRADING:
			raise TradeOnWrongState('lock', self.state)
		if self.locked[1]:
			raise TypeError("Can not lock a trade that is already locked by the client.")

		await self.client.main.send(Packet.new(31, 9).writeBool(True))

	async def unlock(self):
		"""|coro|
		Unlocks (cancels the confirmation) the trade."""
		if self.state != TradeState.TRADING:
			raise TradeOnWrongState('lock', self.state)
		if not self.locked[1]:
			raise TypeError("Can not unlock a trade that is not locked by the client.")

		await self.client.main.send(Packet.new(31, 9).writeBool(False))



import asyncio


class TFMProtocol(asyncio.Protocol):
	def __init__(self, conn):
		self.buffer = bytearray()
		self.client = conn.client
		self.connection = conn
		self.length = 0

	def data_received(self, data):
		self.buffer.extend(data)

		while len(self.buffer) > self.length:
			if self.length == 0:
				for i in range(5):
					byte = self.buffer.pop(0)
					self.length |= (byte & 127) << (i * 7)

					if not byte & 0x80:
						break
				else:
					raise Exception("wtf")

			if len(self.buffer) >= self.length:
				self.client.data_received(bytes(self.buffer[:self.length]), self.connection)
				del self.buffer[:self.length]
				self.length = 0

	def connection_made(self, transport):
		self.connection.open = True
		self.client.dispatch('connection_made', self.connection)

	def connection_lost(self, exc):
		self.connection.open = False

		if exc is not None:
			future = self.client.dispatch('connection_error', self.connection, exc)
			if future is not None:
				# This future is a wrapper for _run_event, which returns True if the event
				# ran successfully, False otherwise.
				# If it returns False, it already handled the auto_restart.
				if not asyncio.ensure_future(future):
					return

			if self.client.auto_restart:
				self.client.loop.create_task(self.client.restart_soon())
			else:
				self.client.close()


class Connection:
	"""Represents the connection between the client and the host."""
	PROTOCOL = TFMProtocol

	def __init__(self, name, client, loop):
		self.name = name
		self.client = client
		self.loop = loop

		self.address = None
		self.protocol = None
		self.transport = None

		self.fingerprint = 0
		self.open = False

	def _factory(self):
		return Connection.PROTOCOL(self)

	async def connect(self, host, port):
		"""|coro|
		Connect the client to the host:port
		"""
		self.address = (host, port)
		self.transport, self.protocol = await self.loop.create_connection(self._factory, host, port)

	async def send(self, packet, cipher=False):
		"""|coro|
		Send a packet to the socket

		:param packet: :class:`aiotfm.Packet` the packet to send.
		:param cipher: :class:`bool` whether or not the packet should be ciphered before sending it.
		"""
		if not self.open:
			return

		if cipher:
			packet.xor_cipher(self.client.keys.msg, self.fingerprint)

		self.transport.write(packet.export(self.fingerprint))
		self.fingerprint = (self.fingerprint + 1) % 100

	def close(self):
		"""Closes the connection."""
		self.open = False
		if not self.transport.is_closing():
			self.transport.write_eof()
			self.transport.close()

	def abort(self):
		"""Abort the connection."""
		self.transport.abort()
		self.close()


import aiohttp

from aiotfm import __version__
from aiotfm.errors import EndpointError, InternalError, MaintenanceError


class Keys:
	"""Represents the keys used by the client to communicate to the server."""
	def __init__(self, **keys):
		self.auth = keys.pop('auth_key', 0)
		self.connection = keys.pop('connection_key', '')
		self.identification = keys.pop('identification_keys', [])
		self.msg = [k & 0xff for k in keys.pop('msg_keys', [])]
		self.packet = keys.pop('packet_keys', [])
		self.version = keys.pop('version', 0)
		self.kwargs = keys


async def get_keys(tfm_id, token):
	"""|coro|
	Fetch the keys required to log into the game.

	:param tfm_id: :class:`int` your Transformice user id.
	:param token: :class:`str` your api token.
	"""
	url = 'https://api.tocu.tk/get_transformice_keys.php'
	params = {'tfmid': tfm_id, 'token': token}
	headers = {"User-Agent": f"Mozilla/5.0 aiotfm/{__version__}"}

	async with aiohttp.ClientSession() as session:
		async with session.get(url, params=params, headers=headers) as resp:
			data = await resp.json()

	success = data.pop('success', False)
	internal_error = data.pop('internal_error', True)
	internal_error_step = data.pop('internal_error_step', 0)
	error = data.pop('error', None)

	if success:
		if not internal_error:
			keys = Keys(**data)
			if len(keys.packet) > 0 and len(keys.identification) > 0 and len(keys.msg) > 0:
				return keys

			raise EndpointError('Something goes wrong: A key is empty ! {}'.format(data))

		if internal_error_step == 2:
			raise MaintenanceError('The game might be in maintenance mode.')

		message = 'An internal error occur: {}'.format(internal_error_step)
		raise InternalError(message)

	raise EndpointError("Can't get the keys. Error info: {}".format(error))


from aiotfm.utils.shakikoo import shakikoo
from aiotfm.utils.date import Date
from aiotfm.utils.get_keys import get_keys
from aiotfm.utils.locale import Translation, Locale

__all__ = ['shakikoo', 'Date', 'get_keys', 'Translation', 'Locale']


import asyncio
import random
import sys
import time
import traceback
import warnings

from aiotfm.packet import Packet
from aiotfm.utils import Locale, get_keys, shakikoo
from aiotfm.connection import Connection
from aiotfm.player import Profile, Player
from aiotfm.tribe import Tribe
from aiotfm.message import Message, Whisper, Channel, ChannelMessage
from aiotfm.shop import Shop
from aiotfm.inventory import Inventory, InventoryItem, Trade
from aiotfm.room import Room
from aiotfm.enums import TradeError, Community
from aiotfm.errors import AiotfmException, InvalidEvent, CommunityPlatformError, \
	AlreadyConnected, IncorrectPassword, LoginError, ServerUnreachable


class Client:
	"""Represents a client that connects to Transformice.
	Two argument can be passed to the :class:`Client`.

	.. _event loop: https://docs.python.org/3/library/asyncio-eventloops.html

	Parameters
	----------
	community: Optional[:class:`int`]
		Defines the community of the client. Defaults to 0 (EN community).
	auto_restart: Optional[:class:`bool`]
		Whether the client should automatically restart on error. Defaults to False.
	loop: Optional[event loop]
		The `event loop`_ to use for asynchronous operations. If ``None`` is passed (defaults),
		the event loop used will be ``asyncio.get_event_loop()``.

	Attributes
	----------
	username: Optional[:class:`str`]
		The bot's username received from the server. Might be None if the bot didn't log in yet.
	room: Optional[:class:`aiotfm.room.Room`]
		The bot's room. Might be None if the bot didn't log in yet or couldn't join any room yet.
	trade: Optional[:class:`aiotfm.inventory.Trade`]
		The current trade that's going on (i.e: both traders accepted it).
	trades: :class:`list`[:class:`aiotfm.inventory.Trade`]
		All the trades that the bot participates. Most of them might be invitations only.
	inventory: Optional[:class:`aiotfm.inventory.Inventory`]
		The bot's inventory. Might be None if the bot didn't log in yet or it didn't receive
		anything.
	locale: :class:`aiotfm.locale.Locale`
		The bot's locale (translations).
	"""
	LOG_UNHANDLED_PACKETS = False

	def __init__(self, community=Community.en, auto_restart=False, loop=None):
		self.loop = loop or asyncio.get_event_loop()

		self.main = Connection('main', self, self.loop)
		self.bulle = None

		self._waiters = {}

		self.room = None
		self.trade = None
		self.trades = {}
		self.inventory = None

		self.username = None
		self.locale = Locale()
		self.community = Community(community)
		self.cp_fingerprint = 0

		self.keys = None
		self.authkey = 0

		self.auto_restart = auto_restart
		self.api_tfmid = None
		self.api_token = None

		self._channels = []

	def data_received(self, data, connection):
		"""|coro|
		Dispatches the received data.

		:param data: :class:`bytes` the received data.
		:param connection: :class:`aiotfm.connection.Connection` the connection that received
			the data.
		"""
		self.dispatch('raw_socket', connection, Packet(data))
		try:
			self.loop.create_task(self.handle_packet(connection, Packet(data)))
		except Exception:
			traceback.print_exc()

	async def handle_packet(self, connection: Connection, packet: Packet):
		"""|coro|
		Handles the known packets and dispatches events.
		Subclasses should handle only the unhandled packets from this method.

		Example: ::
			class Bot(aiotfm.Client):
				async def handle_packet(self, conn, packet):
					handled = await super().handle_packet(conn, packet.copy())

					if not handled:
						# Handle here the unhandled packets.
						pass

		:param connection: :class:`aiotfm.connection.Connection` the connection that received
			the packet.
		:param packet: :class:`aiotfm.Packet` the packet.
		:return: True if the packet got handled, False otherwise.
		"""
		CCC = packet.readCode()
		if CCC == (1, 1): # Old packets
			oldCCC, *data = packet.readString().split(b'\x01')
			data = list(map(bytes.decode, data))
			oldCCC = tuple(oldCCC[:2])

			self.dispatch('old_packet', connection, oldCCC, data)
			return await self.handle_old_packet(connection, oldCCC, data)

		if CCC == (5, 21): # Joined room
			self.room = room = Room(private=not packet.readBool(), name=packet.readUTF())
			self.dispatch('joined_room', room)

		elif CCC == (6, 6): # Room message
			player_id = packet.read32()
			username = packet.readUTF()
			commu = packet.read8()
			message = packet.readUTF()
			player = self.room.get_player(pid=player_id)

			if player is None:
				player = Player(username, pid=player_id)

			self.dispatch('room_message', Message(player, message, commu, self))

		elif CCC == (6, 20): # Server message
			packet.readBool() # if False then the message will appear in the #Server channel
			t_key = packet.readUTF()
			t_args = [packet.readUTF() for i in range(packet.read8())]
			self.dispatch('server_message', self.locale[t_key], *t_args)

		elif CCC == (8, 5): # Show emoji
			player = self.room.get_player(pid=packet.read32())
			emoji = packet.read8()
			self.dispatch('emoji', player, emoji)

		elif CCC == (8, 16): # Profile
			self.dispatch('profile', Profile(packet))

		elif CCC == (8, 20): # Shop
			self.dispatch('shop', Shop(packet))

		elif CCC == (8, 22): # Skills
			skills = {}
			for _ in range(packet.read8()):
				key, value = packet.read8(), packet.read8()
				skills[key] = value
			self.dispatch('skills', skills)

		elif CCC == (16, 2): # Tribe invitation received
			author = packet.readUTF()
			tribe = packet.readUTF()
			self.dispatch('tribe_inv', author, tribe)

		elif CCC == (26, 2): # Logged in successfully
			player_id = packet.read32()
			self.username = username = packet.readUTF()
			played_time = packet.read32()
			community = Community(packet.read8())
			pid = packet.read32()
			self.dispatch('logged', player_id, username, played_time, community, pid)

		elif CCC == (26, 3): # Handshake OK
			online_players = packet.read32()
			connection.fingerprint = packet.read8()
			community = Community[packet.readUTF()]
			country = packet.readUTF()
			self.authkey = packet.read32()

			self.loop.create_task(self._heartbeat_loop())

			await connection.send(Packet.new(8, 2).write8(self.community.value).write8(0))

			os_info = Packet.new(28, 17).writeString('en').writeString('Linux')
			os_info.writeString('LNX 29,0,0,140').write8(0)

			await connection.send(os_info)
			self.dispatch('login_ready', online_players, community, country)

		elif CCC == (26, 12): # Login result
			self.dispatch('login_result', packet.read8(), packet.readUTF(), packet.readUTF())

		elif CCC == (26, 25): # Ping
			self.dispatch('ping')

		elif CCC == (28, 6): # Server ping
			await connection.send(Packet.new(28, 6).write8(packet.read8()))

		elif CCC == (29, 6): # Lua logs
			self.dispatch('lua_log', packet.readUTF())

		elif CCC == (31, 1): # Inventory data
			self.inventory = Inventory.from_packet(packet)
			self.inventory.client = self
			self.dispatch('inventory_update', self.inventory)

		elif CCC == (31, 2): # Update inventory item
			item_id = packet.read16()
			quantity = packet.read8()

			if item_id in self.inventory.items:
				item = self.inventory.items[item_id]
				previous = item.quantity
				item.quantity = quantity
				self.dispatch('item_update', item, previous)

			else:
				item = InventoryItem(item_id=item_id, quantity=quantity)
				self.inventory.items[item.id] = item
				self.dispatch('new_item', item)

		elif CCC == (31, 5): # Trade invite
			pid = packet.read32()

			self.trades[pid] = Trade(self, self.room.get_player(pid=pid))
			self.dispatch('trade_invite', self.trades[pid])

		elif CCC == (31, 6): # Trade error
			name = packet.readUTF().lower()
			error = packet.read8()

			if name == self.username.lower():
				trade = self.trade
			else:
				for t in self.trades.values():
					if t.trader.lower() == name:
						trade = t
						break

			self.dispatch('trade_error', trade, TradeError[error])
			trade._close()

		elif CCC == (31, 7): # Trade start
			pid = packet.read32()
			trade = self.trades.get(pid)

			if trade is None:
				raise AiotfmException(f'Cannot find the trade from pid {pid}.')

			trade._start()
			self.trade = trade
			self.dispatch('trade_start')

		elif CCC == (31, 8): # Trade items
			export = packet.readBool()
			id_ = packet.read16()
			quantity = (1 if packet.readBool() else -1) * packet.read8()

			items = self.trade.exports if export else self.trade.imports
			items.add(id_, quantity)

			trader = self if export else self.trade.trader
			self.trade.locked = [False, False]
			self.dispatch('trade_item_change', trader, id_, quantity, items.get(id_))

		elif CCC == (31, 9): # Trade lock
			index = packet.read8()
			locked = packet.readBool()
			if index > 1:
				self.trade.locked = [locked, locked]
				who = "both"
			else:
				self.trade.locked[index] = locked
				who = self.trade.trader if index == 0 else self

			self.dispatch('trade_lock', who, locked)

		elif CCC == (31, 10): # Trade complete
			trade = self.trade
			self.trade._close(succeed=True)

		elif CCC == (44, 1): # Bulle switching
			timestamp = packet.read32()
			player_id = packet.read32()
			bulle_ip = packet.readUTF()
			ports = packet.readUTF().split('-')

			if self.bulle is not None:
				self.bulle.close()

			self.bulle = Connection('bulle', self, self.loop)
			await self.bulle.connect(bulle_ip, random.choice(ports))
			await self.bulle.send(Packet.new(*CCC).write32(timestamp).write32(player_id))

		elif CCC == (44, 22): # Fingerprint offset changed
			connection.fingerprint = packet.read8()

		elif CCC == (60, 3): # Community platform
			TC = packet.read16()
			self.dispatch('raw_cp', TC, packet.copy())

			if TC == 3: # Connected to the community platform
				self.dispatch('ready')

			elif TC == 55: # Channel join result
				result = packet.read8()
				self.dispatch('channel_joined_result', result)

			elif TC == 57: # Channel leave result
				result = packet.read8()
				self.dispatch('channel_leaved_result', result)

			elif TC == 59: # Channel /who result
				idSequence = packet.read32()
				result = packet.read8()
				players = [Player(packet.readUTF()) for _ in range(packet.read16())]
				self.dispatch('channel_who', idSequence, players)

			elif TC == 62: # Joined a channel
				name = packet.readUTF()

				if name in self._channels:
					channel = [c for c in self._channels if c == name][0]
				else:
					channel = Channel(name, self)
					self._channels.append(channel)

				self.dispatch('channel_joined', channel)

			elif TC == 63: # Quit a channel
				name = packet.readUTF()
				if name in self._channels:
					self._channels.remove(name)

				self.dispatch('channel_closed', name)

			elif TC == 64: # Channel message
				author, community = packet.readUTF(), packet.read32()
				channel_name, message = packet.readUTF(), packet.readUTF()
				channel = self.get_channel(channel_name)

				if channel is None:
					channel = Channel(channel_name, self)
					self._channels.append(channel)

				channel_message = ChannelMessage(author, community, message, channel)
				self.dispatch('channel_message', channel_message)

			elif TC == 65: # Tribe message
				author, message = packet.readUTF(), packet.readUTF()
				self.dispatch('tribe_message', author, message)

			elif TC == 66: # Whisper
				author = Player(packet.readUTF())
				commu = packet.read32()
				receiver = Player(packet.readUTF())
				message = packet.readUTF()

				author = self.room.get_player(name=author, default=author)
				receiver = self.room.get_player(name=receiver, default=receiver)
				self.dispatch('whisper', Whisper(author, commu, receiver, message, self))

			elif TC == 88: # tribe member connected
				self.dispatch('member_connected', packet.readUTF())

			elif TC == 90: # tribe member disconnected
				self.dispatch('member_disconnected', packet.readUTF())

			else:
				if self.LOG_UNHANDLED_PACKETS:
					print(CCC, TC, bytes(packet.buffer)[4:])
				return False

		elif CCC == (100, 67): # New inventory item
			slot = packet.read8()
			slot = None if slot == 0 else slot
			item_id = packet.read16()
			quantity = packet.read8()

			item = InventoryItem(item_id=item_id, quantity=quantity, slot=slot)
			self.inventory[item_id] = item
			self.dispatch('new_item', item)

		elif CCC == (144, 1): # Set player list
			before = self.room.players
			self.room.players = {}

			for _ in range(packet.read16()):
				player = Player.from_packet(packet)
				self.room.players[player.pid] = player

			self.dispatch('bulk_player_update', before, self.room.players)

		elif CCC == (144, 2): # Add a player
			after = Player.from_packet(packet)
			before = self.room.players.pop(after.pid, None)

			self.room.players[after.pid] = after
			if before is None:
				self.dispatch('player_join', after)
			else:
				self.dispatch('player_update', before, after)

		else:
			if self.LOG_UNHANDLED_PACKETS:
				print(CCC, bytes(packet.buffer)[2:])
			return False

		return True

	async def handle_old_packet(self, connection: Connection, oldCCC: tuple, data: list):
		"""|coro|
		Handles the known packets from the old protocol and dispatches events.
		Subclasses should handle only the unhandled packets from this method.

		Example: ::
			class Bot(aiotfm.Client):
				async def handle_old_packet(self, conn, oldCCC, data):
					handled = await super().handle_old_packet(conn, data.copy())

					if not handled:
						# Handle here the unhandled packets.
						pass

		:param connection: :class:`aiotfm.connection.Connection` the connection that received
			the packet.
		:param oldCCC: :class:`tuple` the packet identifiers on the old protocol.
		:param data: :class:`list` the packet data.
		:return: True if the packet got handled, False otherwise.
		"""
		if oldCCC == (8, 7): # Remove a player
			player = self.room.players.pop(int(data[0]), None)

			if player is not None:
				self.dispatch('player_remove', player)

		else:
			if self.LOG_UNHANDLED_PACKETS:
				print("[OLD]", oldCCC, data)
			return False

		return True

	async def _heartbeat_loop(self):
		"""|coro|
		Send a packet every fifteen seconds to stay connected to the game.
		"""
		last_heartbeat = 0
		while self.main.open:
			if self.loop.time() - last_heartbeat >= 15:
				t = time.perf_counter()
				await self.main.send(Packet.new(26, 26))
				if self.bulle is not None and self.bulle.open:
					await self.bulle.send(Packet.new(26, 26))

				self.dispatch('heartbeat', (time.perf_counter() - t) * 1000)
				last_heartbeat = self.loop.time()
			await asyncio.sleep(.5)

	def get_channel(self, name):
		"""Returns a channel from it's name or None if not found.
		:param name: :class:`str` the name of the channel.
		:return: :class:`aiotfm.messages.ChannelMessage` or None
		"""
		if name is None:
			return None

		for channel in self._channels:
			if channel.name == name:
				return channel

	def get_trade(self, player):
		"""Returns the pending/current trade with a player.
		:param player: :class:`aiotfm.room.Player` or :class:`str` the player.
		:return: :class:`aiotfm.trade.Trade` the trade with the player.
		"""
		if not isinstance(player, (str, Player)):
			raise TypeError(f"Expected Player or str types got {type(player)}")

		if isinstance(player, Player):
			return self.trades.get(player.pid)

		player = player.lower()
		for trade in self.trades.values():
			if trade.trader.lower() == player:
				return trade

	def event(self, coro):
		"""A decorator that registers an event.

		More about events [here](Events.md).
		"""
		name = coro.__name__
		if not name.startswith('on_'):
			raise InvalidEvent("'{}' isn't a correct event naming.".format(name))
		if not asyncio.iscoroutinefunction(coro):
			message = "Couldn't register a non-coroutine function for the event {}.".format(name)
			raise InvalidEvent(message)

		setattr(self, name, coro)
		return coro

	def wait_for(self, event, condition=None, timeout=None, stopPropagation=False):
		"""Wait for an event.

		Example: ::
			@client.event
			async def on_room_message(author, message):
				if message == 'id':
					await client.sendCommand('profile '+author)
					profile = await client.wait_for('on_profile', lambda p: p.username == author)
					await client.sendRoomMessage('Your id: {}'.format(profile.id))

		:param event: :class:`str` the event name.
		:param condition: Optionnal[:class:`function`] A predicate to check what to wait for.
			The arguments must meet the parameters of the event being waited for.
		:param timeout: Optionnal[:class:`int`] the number of seconds before
			throwing asyncio.TimeoutError
		:return: :class:`asyncio.Future` a future that you must await.
		"""
		event = event.lower()
		future = self.loop.create_future()

		if condition is None:
			def everything(*a):
				return True
			condition = everything

		if event not in self._waiters:
			self._waiters[event] = []

		self._waiters[event].append((condition, future, stopPropagation))

		return asyncio.wait_for(future, timeout)

	async def _run_event(self, coro, event_name, *args, **kwargs):
		"""|coro|
		Runs an event and handle the error if any.

		:param coro: a coroutine function.
		:param event_name: :class:`str` the event's name.
		:param args: arguments to pass to the coro.
		:param kwargs: keyword arguments to pass to the coro.

		:return: :class:`bool` whether the event ran successfully or not
		"""
		try:
			await coro(*args, **kwargs)
			return True
		# except asyncio.CancelledError:
		# 	raise
		except Exception as e:
			if hasattr(self, 'on_error'):
				try:
					await self.on_error(event_name, e, *args, **kwargs)
				# except asyncio.CancelledError:
				# 	raise
				except Exception:
					if not self.auto_restart:
						self.close()
				finally:
					if self.auto_restart:
						asyncio.ensure_future(self.restart_soon(), loop=self.loop)
		return False

	def dispatch(self, event, *args, **kwargs):
		"""Dispatches events

		:param event: :class:`str` event's name. (without 'on_')
		:param args: arguments to pass to the coro.
		:param kwargs: keyword arguments to pass to the coro.

		:return: :class:`Task` the _run_event wrapper task
		"""
		method = 'on_' + event

		if method in self._waiters:
			to_remove = []
			waiters = self._waiters[method]
			for i, (cond, fut, stop) in enumerate(waiters):
				if fut.cancelled():
					to_remove.append(i)
					continue

				try:
					result = bool(cond(*args))
				except Exception as e:
					fut.set_exception(e)
				else:
					if result:
						fut.set_result(args[0] if len(args) == 1 else args if len(args) > 0 else None)
						if stop:
							del waiters[i]
							return None
						to_remove.append(i)

			if len(to_remove) == len(waiters):
				del self._waiters[method]
			else:
				for i in to_remove[::-1]:
					del waiters[i]

		coro = getattr(self, method, None)
		if coro is not None:
			dispatch = self._run_event(coro, method, *args, **kwargs)
			return asyncio.ensure_future(dispatch, loop=self.loop)

	async def on_error(self, event, err, *a, **kw):
		"""Default on_error event handler. Prints the traceback of the error."""
		message = '\nAn error occurred while dispatching the event "{0}":\n\n{2}'
		tb = traceback.format_exc(limit=-3)
		print(message.format(event, err, tb), file=sys.stderr)
		return message.format(event, err, tb)

	async def on_connection_error(self, conn, error):
		"""Default on_connection_error event handler. Prints the error."""
		print('{0.__class__.__name__}: {0}'.format(error), file=sys.stderr)

	async def on_login_result(self, code, *args):
		"""Default on_login_result handler. Raise an error and closes the connection."""
		self.loop.call_later(5, self.close)
		if code == 1:
			raise AlreadyConnected()
		if code == 2:
			raise IncorrectPassword()
		raise LoginError(code)

	async def connect(self):
		"""|coro|
		Creates a connection with the main server.
		"""

		for port in random.sample([13801, 11801, 12801, 14801], 4):
			try:
				await self.main.connect('94.23.193.229', port)
			except Exception:
				pass
			else:
				break
		else:
			raise ServerUnreachable('Unable to connect to the server.')

		while not self.main.open:
			await asyncio.sleep(.1)

	async def sendHandshake(self):
		"""|coro|
		Sends the handshake packet so the server recognizes this socket as a player.
		"""
		packet = Packet.new(28, 1).write16(self.keys.version).writeString(self.keys.connection)
		packet.writeString('Desktop').writeString('-').write32(0x1fbd).writeString('')
		packet.writeString('74696720697320676f6e6e61206b696c6c206d7920626f742e20736f20736164')
		packet.writeString(
			"A=t&SA=t&SV=t&EV=t&MP3=t&AE=t&VE=t&ACC=t&PR=t&SP=f&SB=f&DEB=f&V=LNX 29,0,0,140&M=Adobe"
			" Linux&R=1920x1080&COL=color&AR=1.0&OS=Linux&ARCH=x86&L=en&IME=t&PR32=t&PR64=t&LS=en-U"
			"S&PT=Desktop&AVD=f&LFD=f&WD=f&TLS=t&ML=5.1&DP=72")
		packet.write32(0).write32(0x6257).writeString('')

		await self.main.send(packet)

	async def start(self, api_tfmid, api_token, keys=None):
		"""|coro|
		Starts the client.

		:param api_tfmid: :class:`int` or :class:`str` your Transformice id.
		:param api_token: :class:`str` your token to access the API.
		"""
		if keys is not None:
			self.keys = keys
		else:
			self.api_tfmid = api_tfmid
			self.api_token = api_token
			self.keys = await get_keys(api_tfmid, api_token)

		await self.connect()
		await self.sendHandshake()
		await self.locale.load()

	async def restart_soon(self, *args, delay=5.0, **kwargs):
		"""Restarts the client in several seconds.
		:param delay: :class:`int` the delay before restarting. Default is 5 seconds.
		:param args: arguments to pass to the :meth:`restart` method.
		:param kwargs: keyword arguments to pass to the :meth:`restart` method."""
		await asyncio.sleep(delay)
		await self.restart(*args, **kwargs)

	async def restart(self, keys=None):
		"""Restarts the client.
		:param keys:"""
		self.dispatch("restart")

		self.close()

		if keys is not None:
			self.keys = keys
		else:
			self.keys = keys = await get_keys(self.api_tfmid, self.api_token)

		await self.connect()
		await self.sendHandshake()
		await self.locale.load()

	async def login(self, username, password, encrypted=True, room='1'):
		"""|coro|
		Log in the game.

		:param username: :class:`str` the client username.
		:param password: :class:`str` the client password.
		:param encrypted: Optional[:class:`bool`] whether the password is already encrypted or not.
		:param room: Optional[:class:`str`] the room where the client will be logged in.
		"""
		if not encrypted:
			password = shakikoo(password)

		packet = Packet.new(26, 8).writeString(username).writeString(password)
		packet.writeString("app:/TransformiceAIR.swf/[[DYNAMIC]]/2/[[DYNAMIC]]/4")
		packet.writeString(room).write32(self.authkey ^ self.keys.auth).write8(0).writeString('')
		packet.cipher(self.keys.identification).write8(0)

		await self.main.send(packet)

	def run(self, api_tfmid, api_token, username, password, **kwargs):
		"""A blocking call that do the event loop initialization for you.

		Equivalent to ::
			@bot.event
			async def on_login_ready(*a):
				await bot.login(username, password)

			loop = asyncio.get_event_loop()
			loop.create_task(bot.start(api_id, api_token))
			loop.run_forever()
		"""
		if self.auto_restart:
			self.auto_restart = False
			warnings.warn(
				"The usage of Client.run and Client.auto_restart is not allowed. "
				"Use Client.start method to enable automatic restart."
			)

		keys = kwargs.pop('keys', None)
		asyncio.ensure_future(self.start(api_tfmid, api_token, keys=keys), loop=self.loop)
		self.loop.run_until_complete(self.wait_for('on_login_ready'))
		asyncio.ensure_future(self.login(username, password, **kwargs), loop=self.loop)

		self.loop.run_forever()
		# try:
		# 	self.loop.run_forever()
		# except:
		# 	# add self.close
		# 	# asyncio.ensure_future(self.close())
		# 	raise

	def close(self):
		"""Closes the sockets."""
		self.main.close()
		if self.bulle is not None:
			self.bulle.close()

		if not self.auto_restart:
			# The process is not exited if the loop is still running
			self.loop.close()

	async def sendCP(self, code, data=b''):
		"""|coro|
		Send a packet to the community platform.

		:param code: :class:`int` the community platform code.
		:param data: :class:`Packet` or :class:`bytes` the data.
		"""
		self.cp_fingerprint = fp = (self.cp_fingerprint + 1) % 0XFFFFFFFF

		packet = Packet.new(60, 3).write16(code)
		packet.write32(self.cp_fingerprint).writeBytes(data)
		await self.main.send(packet, cipher=True)

		return fp

	async def sendRoomMessage(self, message):
		"""|coro|
		Send a message to the room.

		:param message: :class:`str` the content of the message.
		"""
		packet = Packet.new(6, 6).writeString(message)

		await self.bulle.send(packet, cipher=True)

	async def sendTribeMessage(self, message):
		"""|coro|
		Send a message to the tribe.

		:param message: :class:`str` the content of the message.
		"""
		await self.sendCP(50, Packet().writeString(message))

	async def sendChannelMessage(self, channel, message):
		"""|coro|
		Send a message to a public channel.

		:param channel: :class:`str` the channel's name.
		:param message: :class:`str` the content of the message.
		"""
		if isinstance(channel, Channel):
			channel = channel.name

		return await self.sendCP(48, Packet().writeString(channel).writeString(message))

	async def whisper(self, username, message, overflow=False):
		"""|coro|
		Whisper to a player.

		:param username: :class:`str` the player to whisper.
		:param message: :class:`str` the content of the whisper.
		:param overflow: :class:`bool` will send the complete message if True, splitted
			in several messages.
		"""
		if isinstance(username, Player):
			username = username.username

		async def send(msg):
			await self.sendCP(52, Packet().writeString(username).writeString(msg))

		if isinstance(message, str):
			message = message.encode()
		message = message.replace(b'<', b'&lt;').replace(b'>', b'&gt;')

		await send(message[:255])
		for i in range(255, len(message), 255):
			await asyncio.sleep(1)
			await self.whisper(username, message[i:i + 255])

	async def getTribe(self, disconnected=True):
		"""|coro|
		Gets the client's :class:`Tribe` and return it

		:param disconnected: :class:`bool` if True retrieves also the disconnected members.
		:return: :class:`Tribe` or ``None``.
		"""
		sid = self.cp_fingerprint + 1
		await self.sendCP(108, Packet().writeBool(disconnected))

		def is_tribe(tc, packet):
			return (tc == 109 and packet.read32() == sid) or tc == 130

		tc, packet = await self.wait_for('on_raw_cp', is_tribe)
		if tc == 109:
			result = packet.readByte()
			if result == 1:
				tc, packet = await self.wait_for('on_raw_cp', lambda tc, p: tc == 130)
			elif result == 17:
				return None
			else:
				raise CommunityPlatformError(118, result)
		return Tribe(packet)

	async def playEmote(self, emote, flag='be'):
		"""|coro|
		Play an emote.

		:param emote: :class:`int` the emote's id.
		:param flag: Optional[:class:`str`] the flag for the emote id 10. Defaults to 'be'.
		"""
		packet = Packet.new(8, 1).write8(id).write32(0)
		if emote == 10:
			packet.writeString(flag)

		await self.bulle.send(packet)

	async def sendSmiley(self, smiley):
		"""|coro|
		Makes the client showing a smiley above it's head.

		:param smiley: :class:`int` the smiley's id. (from 0 to 9)
		"""
		if smiley < 0 or smiley > 9:
			raise AiotfmException('Invalid smiley id')

		packet = Packet.new(8, 5).write8(smiley).write32(0)

		await self.bulle.send(packet)

	async def loadLua(self, lua_code):
		"""|coro|
		Load a lua code in the room.

		:param lua_code: :class:`str` or :class:`bytes` the lua code to send.
		"""
		if isinstance(lua_code, str):
			lua_code = lua_code.encode()

		packet = Packet.new(29, 1).write24(len(lua_code)).writeBytes(lua_code)

		await self.bulle.send(packet)

	async def sendCommand(self, command):
		"""|coro|
		Send a command to the game.

		:param command: :class:`str` the command to send.
		"""
		packet = Packet.new(6, 26).writeString(command[:255])

		await self.main.send(packet, cipher=True)

	async def enterTribe(self):
		"""|coro|
		Enter the tribe house
		"""
		await self.main.send(Packet.new(16, 1))

	async def enterTribeHouse(self):
		"""|coro|
		Alias for :meth:`enterTribe`
		"""
		await self.enterTribe()

	async def joinRoom(self, room_name, community=None, auto=False):
		"""|coro|
		Join a room.
		The event 'on_joined_room' is dispatched when the client has successfully joined the room.

		:param room_name: :class:`str` the room's name.
		:param community: Optional[:class:`int`] the room's community.
		:param auto: Optional[:class:`bool`] joins a random room (I think).
		"""
		packet = Packet.new(5, 38).write8(Community(community or self.community).value)
		packet.writeString(room_name).writeBool(auto)
		await self.main.send(packet)

	async def joinChannel(self, name, permanent=True):
		"""|coro|
		Join a #channel.
		The event 'on_channel_joined' is dispatched when the client has successfully joined
		a channel.

		:param name: :class:`str` the channel's name
		:param permanent: Optional[:class:`bool`] if True (default) the server will automatically
		reconnect the user to this channel when logged in.
		"""
		await self.sendCP(54, Packet().writeString(name).writeBool(permanent))

	async def leaveChannel(self, channel):
		"""|coro|
		Leaves a #channel.

		:param channel: :class:`aiotfm.Channel` channel to leave.
		"""
		if isinstance(channel, Channel):
			name = channel.name
		else:
			name = channel

		await self.sendCP(56, Packet().writeString(name))

	async def enterInvTribeHouse(self, author):
		"""|coro|
		Join the tribe house of another player after receiving an /inv.

		:param author: :class:`str` the author's username who sent the invitation.
		"""
		await self.main.send(Packet.new(16, 2).writeString(author))

	async def recruit(self, player):
		"""|coro|
		Send a recruit request to a player.

		:param player: :class:`str` the player's username you want to recruit.
		"""
		await self.sendCP(78, Packet().writeString(player))

	async def requestShopList(self):
		"""|coro|
		Send a request to the server to get the shop list."""
		await self.main.send(Packet.new(8, 20))

	async def startTrade(self, player):
		"""|coro|
		Starts a trade with the given player.

		:param player: :class:`aiotfm.player.Player` the player to trade with.
		:return: :class:`aiotfm.inventory.Trade` the resulting trade"""
		if isinstance(player, Player) and player.pid == -1:
			player = player.username

		if isinstance(player, str):
			player = self.room.get_player(username=player)
			if player is None:
				raise AiotfmException("The player must be in your room to start a trade.")

		trade = Trade(self, player)

		self.trades[player.pid] = trade
		await trade.accept()
		return trade

	async def requestInventory(self):
		"""|coro|
		Send a request to the server to get the bot's inventory."""
		await self.main.send(Packet.new(31, 1))


import struct

from aiotfm.errors import XXTEAInvalidPacket, XXTEAInvalidKeys, PacketError


class Packet:
	"""Represents a network packet.

	Parameters
	----------
	buffer: Optional[:class:`bytes`]

	Attributes
	----------
	buffer: :class:`bytearray`
		The content of the packet.
	pos: :class:`int`
		The position inside the buffer.
	"""
	def __init__(self, buffer=None):
		if isinstance(buffer, (bytes, memoryview)):
			self.buffer = memoryview(buffer)
			self.__read = True
			self.__write = False
		else:
			if buffer is not None:
				self.buffer = buffer
			else:
				self.buffer = bytearray()

			self.__read = False
			self.__write = True

		self.pos = 0

	def __repr__(self):
		return '<Packet {!r}>'.format(bytes(self))

	def __bytes__(self):
		if self.__read:
			return self.buffer.obj
		return bytes(self.buffer)

	def __del__(self):
		if isinstance(self.buffer, memoryview):
			self.buffer.release()

	@classmethod
	def new(cls, c, cc=None):
		"""Create a new instance of Packet initialized by two bytes: c and cc."""
		msg = cls()
		if isinstance(c, (tuple, list)):
			c, cc = c
		if cc is None:
			return msg.write16(c)
		return msg.write8(c).write8(cc)

	def copy(self):
		"""Returns a copy of the Packet"""
		if self.__read:
			return Packet(self.buffer[self.pos:])

		p = Packet()
		p.buffer = self.buffer.copy()
		return p

	def readBytes(self, nbr=1):
		"""Read raw bytes from the buffer."""
		if not self.__read:
			raise PacketError('This packet is in write-only mode.')

		self.pos += nbr
		return self.buffer[self.pos - nbr:self.pos]

	def readCode(self):
		"""Read two bytes: c and cc."""
		return self.read8(), self.read8()

	def read8(self):
		"""Read a single byte from the buffer."""
		if not self.__read:
			raise PacketError('This packet is in write-only mode.')

		self.pos += 1
		return self.buffer[self.pos - 1]

	def read16(self):
		"""Read a short (two bytes) from the buffer"""
		return struct.unpack('>H', self.readBytes(2))[0]

	def read24(self):
		"""Read three bytes from the buffer"""
		return int.from_bytes(self.readBytes(3), 'big')

	def read32(self):
		"""Read an int (four bytes) from the buffer"""
		return struct.unpack('>I', self.readBytes(4))[0]

	def readBool(self):
		"""Read a boolean (one byte) from the buffer"""
		return self.read8() == 1

	def readString(self):
		"""return a encoded string (in bytes)"""
		return bytes(self.readBytes(self.read16()))

	def readUTF(self):
		"""return a decoded string"""
		return self.readString().decode()

	def writeBytes(self, content):
		"""Write raw bytes to the buffer"""
		if not self.__write:
			raise PacketError('This packet is in read-only mode.')

		if isinstance(content, Packet):
			self.buffer.extend(content.buffer)
		else:
			self.buffer.extend(content)
		return self

	def writeCode(self, c, cc):
		"""Write two bytes: c and cc."""
		return self.write8(c).write8(cc)

	def write8(self, value):
		"""Write a single byte to the buffer"""
		if not self.__write:
			raise PacketError('This packet is in read-only mode.')

		self.buffer.append(value & 0xff)
		return self

	def write16(self, value):
		"""Write a short (two bytes) to the buffer"""
		if not self.__write:
			raise PacketError('This packet is in read-only mode.')

		self.buffer.extend(struct.pack('>H', value & 0xffff))
		return self

	def write24(self, value):
		"""Write three bytes to the buffer"""
		if not self.__write:
			raise PacketError('This packet is in read-only mode.')

		self.buffer.extend((value & 0xffffff).to_bytes(3, 'big'))
		return self

	def write32(self, value):
		"""Write an int (four bytes) to the buffer"""
		if not self.__write:
			raise PacketError('This packet is in read-only mode.')

		self.buffer.extend(struct.pack('>I', value & 0xffffffff))
		return self

	def writeBool(self, value):
		"""Write a boolean (one byte) to the buffer"""
		return self.write8(1 if value else 0)

	def writeString(self, string):
		"""Write a string to the buffer"""
		if isinstance(string, str):
			string = string.encode()
		return self.write16(len(string)).writeBytes(string)

	def writeUTF(self, string):
		"""Write a string to the buffer. Alias for .writeString"""
		return self.writeString(string)

	def export(self, fp=0):
		"""Generates the header then converts the whole packet to bytes and returns it."""
		m = Packet()
		size = len(self.buffer)
		size_type = size >> 7
		while size_type != 0:
			m.write8(size & 0x7f | 0x80)
			size = size_type
			size_type >>= 7
		m.write8(size & 0x7f)
		m.write8(fp)

		return bytes(m.buffer + self.buffer)

	def xor_cipher(self, key, fp):
		"""Cipher the packet with the XOR algorithm."""
		data = memoryview(self.buffer)[2:]
		self.buffer[2:] = (byte ^ key[i % 20] for i, byte in enumerate(data, fp + 1))
		return self

	def cipher(self, key):
		"""Cipher the packet with the XXTEA algorithm."""
		if len(self.buffer) < 2:
			raise XXTEAInvalidPacket("The Packet is empty.")
		if len(key) < 4:
			raise XXTEAInvalidKeys(str(key))

		header = self.buffer[:2]
		length = len(self.buffer) - 2
		if length % 4 > 0:
			pad = 4 - length % 4
			self.buffer.extend(bytes(pad))
			length += pad

		chunks = struct.unpack(f'>{length//4}I', self.buffer[2:])
		chunks = xxtea_encode(list(chunks), len(chunks), key)

		packet = Packet(header).write16(len(chunks))
		packet.writeBytes(struct.pack(f'>{len(chunks)}I', *chunks))

		self.buffer = packet.buffer
		return self


DELTA = 0x9e3779b9


def xxtea_encode(v, n, key):
	"""https://en.wikipedia.org/wiki/XXTEA"""
	cycles = 6 + 52 // n
	sum_ = 0
	z = v[-1]
	for _ in range(cycles):
		sum_ = (sum_ + DELTA) & 0xffffffff
		e = sum_ >> 2 & 3
		for p in range(n):
			y = v[(p + 1) % n]
			z = v[p] = (v[p] + (
				((z >> 5 ^ y << 2) + (y >> 3 ^ z << 4)) ^ ((sum_ ^ y) + (key[(p & 3) ^ e] ^ z))
			)) & 0xffffffff
	return v


from aiotfm.enums import ChatCommunity
from aiotfm.packet import Packet


class Message:
	"""Represents any message from the chat.
	Convert an instance to string to get the representation in game of the message.

	Attributes
	----------
	author: `aiotfm.Player`
		The message's author.
	community: `aiotfm.enum.ChatCommunity`
		The author's community. Note: the community isn't the author's language!
	content: `str`
		The actual content of the message.
	"""
	def __init__(self, author, content, community, client):
		self.author = author
		self.community = ChatCommunity(community)
		self.content = content
		self._client = client

	def __str__(self):
		return '[{0.author}] {0.content}'.format(self)

	def __repr__(self):
		return '<{.__class__.__name__} {}>'.format(self, ' '.join(
			'='.join((k, repr(v)[:32])) for k, v in vars(self).items()
			if not k.startswith('_')
		))


class Whisper(Message):
	"""Represents a whisper from the chat.
	Inherit from :class:`aiotfm.message.Message`.

	Attributes
	----------
	author: `aiotfm.Player`
		The message's author.
	receiver: `aiotfm.Player`
		The message's addressee.
	community: `aiotfm.enum.ChatCommunity`
		The author's community. Note: the community isn't the author's language!
	content: `str`
		The actual content of the message.
	sent: `bool`
		True if the author is the client.
	"""
	def __init__(self, author, community, receiver, content, client):
		super().__init__(author, content, community, client)
		self.receiver = receiver

		self.sent = self.author == client.username

	def __str__(self):
		direction = '<' if self.sent else '>'
		author = self.receiver if self.sent else self.author
		commu = '' if self.sent else '[{}] '.format(self.community.name)
		return f'{direction} {commu}[{author}] {self.content}'

	async def reply(self, msg):
		"""|coro|
		Reply to the author of the message. Shortcut to :meth:`Client.whisper`.
		:param msg: :class:`str` the message."""
		await self._client.whisper(self.author, msg)


class Channel:
	"""Represents a channel (#chat) in the game.

	Attributes
	----------
	name: `str`
		The actual channel's name.
	"""
	def __init__(self, name, client):
		self.name = name
		self._client = client

	def __repr__(self):
		return '<Channel name={.name}>'.format(self)

	def __eq__(self, other):
		if isinstance(other, str):
			return self.name == other
		return self.name == other.name

	async def send(self, message):
		"""|coro|
		Sends a message to the channel.

		:param message: :class:`str` the content of the message"""
		await self._client.sendChannelMessage(self, message)

	async def leave(self):
		"""|coro|
		Leaves the channel."""
		await self._client.leaveChannel(self)

	async def who(self):
		"""|coro|
		Sends the command /who to the channel and returns the list of players.

		:throws: `asyncio.TimeoutError`
		:return: List[`aiotfm.Player`]"""
		def check(idseq, players):
			return idseq == idSequence

		idSequence = await self._client.sendCP(58, Packet().writeString(self.name))
		_, players = await self._client.wait_for('on_channel_who', check, timeout=3)
		return players


class ChannelMessage(Message):
	"""Represents a message from a :class:`Channel`.

	Attributes
	----------
	channel: `Channel`
		The channel where the message is from.
	author: `aiotfm.Player`
		The message's author.
	community: `aiotfm.enum.ChatCommunity`
		The author's community. Note: the community isn't the author's language!
	content: `str`
		The actual content of the message."""
	def __init__(self, author, community, content, channel):
		super().__init__(author, content, community, channel._client)
		self.channel = channel

	async def reply(self, message):
		"""|coro|
		Sends a message to the channel.

		:param message: :class:`str` the content of the message"""
		await self.channel.send(message)

	def __str__(self):
		return '{0.channel.name} [{0.community.value}] [{0.author}] {0.content}'.format(self)


import base64
import hashlib

SALT = bytes((
	0xf7, 0x1a, 0xa6, 0xde, 0x8f, 0x17, 0x76, 0xa8, 0x03, 0x9d, 0x32, 0xb8, 0xa1, 0x56, 0xb2, 0xa9,
	0x3e, 0xdd, 0x43, 0x9d, 0xc5, 0xdd, 0xce, 0x56, 0xd3, 0xb7, 0xa4, 0x05, 0x4a, 0x0d, 0x08, 0xb0
))


def shakikoo(string):
	"""Encrypt a password with the SHAKikoo algorithm."""
	sha256 = hashlib.sha256(string.encode()) # hash the password with SHA256
	hex256 = sha256.hexdigest().encode() # convert it into hexdecamals
	hex256 += SALT # salt it

	hashed = hashlib.sha256(hex256).digest() # re-hash it
	return base64.b64encode(hashed) # return it in base64 bytes